/*
FILE : ASSIGN2.C
Hasan Timucin OZDEMIR , Dec. 1995

This program tries to multiply the two square matrices on the iPSC/860 MIMD
machine. The dimension of hypercube and the dimension of matrix can be
specified by the user.

*/
#include <cube.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define FALSE 0
#define TRUE  1
#define TYPE_FLOAT 100
#define MOD(a,b) (a-((int)a/b)*b)
#define FLOAT_LEN 4

/* Binary to Reflected Gray Code Conversion */
int btog(int binary)
{ return(binary ^ (binary >> 1));
}/* end of btog() */

/* Reflected Gray Code to Binary Conversion */
int gtob(int gray)
{ int binary = gray;
  while((gray = (gray >>1))!= 0) binary ^= gray;
  return(binary);
}/* end of gtob() */

/* Sequential Matrix Multiplication */
Matrix_Multiply(float *A, float *B, float *C, int m, int n, int p)
/* Matrix A is mxn matrix and Matrix B is nxp matrix and the result matrix
   C is mxp matrix. 
*/
{ int i,j,k;
  for(i=0;i<m;printf(" %d ",i),i++)
     for(j=0;j<p;j++)
       for(k=0,C[i*p+j]=0;k<n;k++) C[i*p+j] = C[i*p+j] + A[i*n+k] * B[k*p+j];
}/* end of Matrix_Multiply() */

Matrix_Add(float *A, float *B, float *C, int m, int n)
{ int i,j;
  for(i=0;i<m;i++)
     for(j=0;j<n;j++)
	C[i*m+j] = A[i*m+j] + B[i*m+j];
}/* end of Matrix_Add() */

/* Returns the Mesh id of the processor my_id in the Hypercube */
void get_my_Mesh_id(int *xid, int *yid, long my_id, int px)
{ int binid;
  binid=gtob(my_id);
  *yid=binid / px;
  *xid=binid % px;
}/* end of get_my_Mesh_id() */

/* Returns the Hypercube id of the processor (x,y) in the Mesh */
get_Hyper_id(int x, int y, int px)
{ return(btog(y*px+x));
}/* end of get_Hyper_id() */

/* Preconditioning will shift row elements of A and column elements of B.
   ith row of matrix A will be shifted i times to west circularly.
   ith column of matrix B will be shifted i times to north circularly.
*/ 
void Preconditioning(int px, int py, int n, 
                     float *a, float *b, int mx, int my,int maxnode)
{ int i,x,y,my_id;
  int xid,yid; 
  int start,end,j,ni,k,l;
  float buf;
  long iam;
  char smsg[20],rmsg[20];
  float *aptr,*bptr;  
  float *aRptr,*bRptr;  
  char *r1msg;
 
  aptr=(float *) calloc(my,sizeof(float));
  bptr=(float *) calloc(mx,sizeof(float));
  iam=mynode();
  for(i=0;i<n-1;i++)
  { get_my_Mesh_id(&xid,&yid,mynode(),px); 
    k=i/my;
    printf("\n node %d : i=%d",mynode(),i);
    if(yid==k)
    { start=i-yid*my; 
      for(l=start+1;l<my;l++) aptr[l-(start+1)]=a[l*mx];
      csend(get_Hyper_id((xid-1+px)%px,yid,px),(char *)aptr,FLOAT_LEN*(my-start),get_Hyper_id((xid-1+px)%px,yid,px),0);
      r1msg = (char *) aptr;
      printf("\n node %d : sends ",mynode());
      crecv(iam,r1msg,FLOAT_LEN*(my-start));
      printf("\n node %d : received ",mynode());
      aRptr=(float *)r1msg;
       /* Shift West a array */
      for(k=start+1;k<my;k++)
         for(l=0;l<(mx-1);l++) a[k*mx+l] = a[k*mx+l+1];
      for(k=start+1;k<my;k++) a[k*mx+l] = aRptr[k-(start+1)];
    }/* end of if */
    else if(yid>k)
         { /* send my elements */
           for(l=0;l<my;l++) aptr[l]=a[l*mx];
           csend(get_Hyper_id((xid-1+px)%px,yid,px),(char *)aptr,FLOAT_LEN*my,get_Hyper_id((xid-1+px)%px,yid,px),0);
      printf("\n node %d : sends ",mynode());
           r1msg = (char *) aptr;
           crecv(iam,r1msg,FLOAT_LEN*my);
      printf("\n node %d : received ",mynode());
           aRptr=(float *)r1msg;
           /* Shift West a array */
           for(k=0;k<my;k++)
             for(l=0;l<(mx-1);l++) a[k*mx+l] = a[k*mx+l+1];
           for(k=0;k<my;k++) a[k*mx+l] = aRptr[k];
         }
  }/* end of for */
  /* printf("\n node %d pass 1. precindition \n",mynode());*/
  for(i=0;i<n-1;i++)
  { get_my_Mesh_id(&xid,&yid,mynode(),px); 
    k=i/mx;
    if(xid==k)
    { start = i - xid* mx ;
      for(k=start+1;k<mx;k++) bptr[k-(start+1)]=b[k];
      csend(get_Hyper_id(xid,(yid-1+py)%py,px)+n,(char *)bptr,FLOAT_LEN*(mx-start),get_Hyper_id(xid,(yid-1+py)%py,px),0);
      r1msg = (char *) bptr;
      crecv(iam+n,r1msg,FLOAT_LEN*(mx-start));
      bRptr=(float *)r1msg;
      /* Shift North b array */
      for(l=0;l<(my-1);l++) 
         for(k=start+1;k<mx;k++) b[l*mx+k] = b[(l+1)*mx+k];
      for(k=start+1;k<mx;k++) b[l*mx+k]=bRptr[k-(start+1)];
    }/* end of if */
    else if(xid>k)
         { for(k=0;k<mx;k++) bptr[k]=b[k];
           csend(get_Hyper_id(xid,(yid-1+py)%py,px)+n,(char *)bptr,FLOAT_LEN*mx,get_Hyper_id(xid,(yid-1+py)%py,px),0);
           r1msg = (char *) bptr;
           crecv(iam+n,r1msg,FLOAT_LEN*mx);
           bRptr=(float *)r1msg;
           /* Shift North b array */
           for(l=0;l<(my-1);l++) 
               for(k=0;k<mx;k++) b[l*mx+k] = b[(l+1)*mx+k];
           for(k=0;k<mx;k++) b[l*mx+k]=bRptr[k];
         }
    /*gsync();*/
  }/* end of for */
 /*  printf("\n node %d pass 2. precindition \n",mynode());*/
  return;
}/* end of Preconditioning() */

/* Preconditioning will shift row elements of A and column elements of B.
   ith row of matrix A will be shifted i times to west circularly.
   ith column of matrix B will be shifted i times to north circularly.
*/ 
void Preconditioning0(int px, int py, int n, 
                     float *a, float *b, int mx, int my,int maxnodes)
{ int i,x,y,my_id;
  int xid,yid; 
  int start,end,j,ni,k;
  float buf;
  char smsg[20],rmsg[20];
  gsync();
  printf("\n node %d : in precond  ",mynode());
  for(i=0;i<n-1;i++)
  { get_my_Mesh_id(&xid,&yid,mynode(),px); 
    k=i/my;
    if(yid>=k)
    { if(yid>k) start=0;
      else  start=i-yid*my+1; 
      for(j=start;j<my;j++)
      { sprintf(smsg,"%d",(int)a[j*mx]); 
        /*printf("\n node (%d,%d) : sending (%d,%d) to node (%d,%d) ",xid,yid,0,j,(xid-1+px)%px,yid);*/
        csend(TYPE_FLOAT,smsg,strlen(smsg)+1,get_Hyper_id((xid-1+px)%px,yid,px),0);
         printf("\n node %d : sends %ld",mynode(),get_Hyper_id((xid-1+px)%px,yid,px));

        /*csend(TYPE_FLOAT,a[j,0],1,get_Hyper_id((xid-1+px)%px,yid,px),0);*/
        crecv(TYPE_FLOAT,rmsg,sizeof(rmsg));
        /*printf("\n node (%d,%d) : received (%d,%d) ",xid,yid,mx-1,j);*/
        for(ni=0;ni<(mx-1);ni++) a[j*mx+ni]=a[j*mx+ni+1];
        a[j*mx+ni]=(float)atoi(rmsg);
      }/* end of for */
    }/* end of if */
    gsync();
  }/* end of for */

  for(i=0;i<n-1;i++)
  { get_my_Mesh_id(&xid,&yid,mynode(),px); 
    k=i/mx;
    if(xid>=k)
    { if(xid>k) start=0;
      else start = i - xid* mx + 1;
      for(j=start;j<mx;j++)
      { sprintf(smsg,"%d",(int)b[j]); 
        csend(TYPE_FLOAT,smsg,strlen(smsg)+1,get_Hyper_id(xid,(yid-1+py)%py,px),0);
        /*csend(TYPE_FLOAT,b[0,j],1,get_Hyper_id(xid,(yid-1+py)%py,px),0);*/
        crecv(TYPE_FLOAT,rmsg,sizeof(rmsg));
        for(ni=0;ni<(my-1);ni++) b[ni*mx+j]=b[(ni+1)*mx+j];
        b[ni*mx+j]=(float)atoi(rmsg);
      }/* end of for */
    }/* end of if */
    gsync();
  }/* end of for */
  return;
}/* end of Preconditioning() */
/* Multiply Locally, shift elements of A to west and shift lements of B to
north circularly.
*/
Parallel_Matrix_Multiply(int px, int py, int n, float *a, float *b, 
                         float *c, int mx, int my)
{ int i,x,y,xid,yid;
  int k,m,l;
  float buf;
  char smsg[20],rmsg[20];
  float *aptr,*bptr;
  float *aRptr,*bRptr;
  char *r1msg;
  aptr=(float *) calloc(my,sizeof(float));
  bptr=(float *) calloc(mx,sizeof(float));
  for(i=0;i<n;i++)
  { 
    for(k=0;k<my;k++)
      for(m=0;m<mx;m++)
       c[k*mx+m] = c[k*mx+m] + a[k*mx+m] * b[k*mx+m];
    
    get_my_Mesh_id(&xid,&yid,mynode(),px);
    if(!mynode()) printf(" p %d ",i);

    for(l=0;l<my;l++) aptr[l]=a[l*mx];
    csend(TYPE_FLOAT+i,(char *)aptr,FLOAT_LEN*my,get_Hyper_id((xid-1+px)%px,yid,px),0);
    /*gsync();*/
    r1msg = (char *) aptr;
    crecv(TYPE_FLOAT+i,r1msg,FLOAT_LEN*my);
    aRptr=(float *)r1msg;
    /* Shift West a array */
    for(k=0;k<my;k++)
      for(l=0;l<(mx-1);l++) a[k*mx+l] = a[k*mx+l+1];
    for(k=0;k<my;k++) a[k*mx+l] = aRptr[k];
    /*gsync();*/

    for(k=0;k<mx;k++) bptr[k]=b[k];
    csend(TYPE_FLOAT+2*i+1,(char *)bptr,FLOAT_LEN*mx,get_Hyper_id(xid,(yid-1+py)%py,px),0);
    /*gsync();*/
    r1msg = (char *) bptr;
    crecv(TYPE_FLOAT+2*i+1,r1msg,FLOAT_LEN*mx);
    bRptr=(float *)r1msg;
    /* Shift North b array */
    for(l=0;l<(my-1);l++) 
        for(k=0;k<mx;k++) b[l*mx+k] = b[(l+1)*mx+k];
    for(k=0;k<mx;k++) b[l*mx+k]=bRptr[k];

    /*gsync();*/

  }/* end of for */
  free(aptr);
  free(bptr);
  return;
}/* end of Parallel_Matrix_Multiply() */

float random1()
/* This routine tries to find out a random value by using the clock value
   and the initial value of div1. 
*/
{ float val;
  float div1,div2,div3,div4;
  val = dclock() + 0.11 * (div1-(int)(div1/11)*11);
  if(val<1000) div1 = 10*(val-(int)val);
  else div1 = fmod(val,1000);
  div2 = ((int)div1)/100;
  div3 = ((int)(div1-div2*100))/10 ;
  div4 = (div1-(div2*100+div3*10)) ;
  div2 = (div4+0.1) * (div2+0.1) * (div3+0.1) * 0.3; 
  /*printf("\n%f--> %f --> %f", val,div1,div2*div1 );*/
  return( fmod(div2*div1,3211) );
}/* end of random() */

produce_random(float *A, float *B, int dim)
/* This routine will produce the random value for each element of the A, B */
{ int i;
  /*printf("\n I am in random \n");*/
  srand((unsigned int)((i+2*i)*A[1]));
  for(i=0;i<dim*dim;i++)
  { A[i] = fmod((float)rand()*0.2*(dim-i),dim*dim);
    B[i] = fmod((float)rand()*0.1*(i-dim),dim*dim);
  }/* end of for */
}/* end of produce_random() */

produce_matrix(float **A, float **B, float **C, int dimen)
/* Allocates memory for matrices and if it can nor find enough space
returns FALSE value. 
*/
{ int retval=FALSE;
  unsigned int dim=(unsigned int)dimen;
  if(*A = (float *) calloc(dim*dim,sizeof(float)))
    { if(*B = (float *) calloc(dim*dim,sizeof(float)))
      { if(*C = (float *) calloc(dim*dim,sizeof(float)))
         { produce_random(*A,*B,dim);
	    retval=TRUE;
      }
      else { free(*A); free(*B);}
    }/* end of if */
    else free(*A);
  }/* end of if */
  return(retval);
}/* end of produce_matrix() */

distribute(float *A, float *B, int n, float *a, float *b, float *c,
           int mx, int my, int px, int py)
/* Distribute the matrix A and B to the appropriate processors. */
{ long iam;
  int i,j,m,k;
  char smsg[80],rmsg[80];
  char *r1msg;
  int xid,yid;
  float *bf2;
  float *bf1;

  bf1=(float *)calloc(mx*my,sizeof(float));
  iam=mynode();
  get_my_Mesh_id(&xid,&yid,iam,px);
  /* distribution of A array */
  if(!iam)
  { for(i=0;i<py;i++)
      for(j=0;j<px;j++)
      { for(m=0;m<my;m++) 
          for(k=0;k<mx;k++) bf1[m*mx+k]=A[(i*my+m)*n+j*mx+k];
        csend(i*px+j,(char *)bf1,FLOAT_LEN*mx*my,get_Hyper_id(j,i,px),0);
      }/*   end of for of for */
  }/* end of if */ 
  /*gsync();*/
  r1msg = (char *)bf1;
  crecv(yid*px+xid,r1msg,FLOAT_LEN*mx*my);
  bf2=(float *)r1msg;
  for(m=0;m<my;m++) 
    for(k=0;k<mx;k++) a[m*mx+k]=bf2[m*mx+k];
  /*gsync();*/

  /* distribution of B array */
  if(!iam)
  { for(i=0;i<py;i++)
      for(j=0;j<px;j++)
      { for(m=0;m<my;m++) 
          for(k=0;k<mx;k++) bf1[m*mx+k]=B[(i*my+m)*n+j*mx+k];
        csend(i*px+j+px*py,(char *)bf1,FLOAT_LEN*mx*my,get_Hyper_id(j,i,px),0);
      }/*   end of for of for */
  }/* end of if */ 
  /*gsync();*/
  r1msg = (char *) bf1;
  crecv(px*yid+xid+px*py,r1msg,FLOAT_LEN*mx*my);
  bf2=(float *) r1msg;;
  for(m=0;m<my;m++) 
    for(k=0;k<mx;k++) b[m*mx+k]=bf2[m*mx+k];
  /*gsync();*/
  free(bf1);

  /* initialization of c array */ 
  for(i=0;i<my;i++)
     for(j=0;j<mx;j++) c[i*mx+j]=0.0;
  return;

}/* end of distribute() */

readvalue(int *MatrixDimen,int *HyperDimen, int *px, int *py)
{  int buf1,buf2;
   int buf3,buf4; 
   int ok=FALSE;
   *MatrixDimen = 8;
   *HyperDimen = nodedim();
   *px = (numnodes())/2;
   *py = (numnodes())/(*px);

   printf("\n HyperCube Dimen %d, nodes %d",nodedim(),numnodes());

   printf("\n Enter Matrix dimension :");
   scanf("%d",&buf1);

   do 
   { ok = TRUE;
     printf("\n Enter Hypercube dimension :");
     scanf("%d",&buf2);
     if(buf2>*HyperDimen) 
     { printf("\n you cannot give more than the possible cube dimension!\n");
       ok = TRUE;
     }/* end of if */
   } while(!ok);
   
   /* calculate px,py  */
   if(!((buf2)%2))  /* Hypercube dimension is even*/ 
	     buf3=buf4=pow(2,buf2/2);
    else { buf3 = pow(2,(buf2-1)/2);
	     buf4 = 2 * buf3;
         }
   printf("\n px = %d, py = %d ",buf3,buf4);
   printf("\n Number of available nodes = %d  \n",numnodes());
   *MatrixDimen = buf1;
   *HyperDimen = buf2;
   *px = buf3; 
   *py = buf4; 

  return;
}/* end of readvalues */

print_matrix(float *a, int mx,int my, int px, int py)
/* Prints the matrix 
   a  : matrix
   mx,my : number of columns, number of rows
   py,px : mesh size, mesh have py rows and px columns.
*/
{ int i,j,k,m;
  int xid,yid;
  get_my_Mesh_id(&xid,&yid,mynode(),px);
  /*printf("\n My Id (%d,%d)",xid,yid);*/

  for(i=0;i<py;i++)
   for(j=0;j<px;j++)
    { if(mynode()==get_Hyper_id(j,i,px))
      { for(printf("\n(%d,%d) : ",i,j),m=0;m<my;m++)
        { printf("\n (%d,%d) ",j*my+m,i*mx); 
          for(k=0;k<mx;k++)
          { if(((k/4)*4)==k) printf("\n (%d,%d) ",j*my+m,i*mx+k); 
             printf(" %10.5f ",a[m*mx+k]);
          }
        }
      }
      /*gsync();*/
    } 
  printf("\n");
}/* end of print_matrix() */

print_matrixS(float *a, int mx,int my)
/* print matrix for sequential purposes
   a : matrix
   mx,my : number of columns, number of rows
*/
{ int i,j,k,m;
  for(m=0;m<my;printf("\n (%d,%d) ",m,0),m++)
     for(k=0;k<mx;k++)
      { if(((k/4)*4)==k) printf("\n (%d,%d) ",m,k);
        printf(" %12.5f ",a[m*mx+k]);
      }
  printf("\n");
}/* end of print_matrixS() */

Fprint_matrixS(FILE *fp,float *a, int mx,int my)
/* print matrix to file for sequential purposes
   fp : output file stream
   a : matrix
   mx,my : number of columns, number of rows
*/
{ int i,j,k,m;
  fprintf(fp,"\n");
  for(m=0;m<my;fprintf(fp,"\n (%d,%d) ",m,0),m++)
     for(k=0;k<mx;k++)
     { if(((k/4)*4)==k) fprintf(fp,"\n (%d,%d) ",m,k);
        fprintf(fp," %12.3f ",a[m*mx+k]);
     }
  fprintf(fp,"\n");
}/* end of Fprint_matrixS() */

main()
{ int MatrixDimen,HyperDimen;
  int px,py;
  int mx,my;
  char smsg[10],rmsg[10];
  int i,j,maxnodes;
  float *A, *B, *C;
  float *sa, *sb, *sc;
  unsigned long start_time,end_time;
  long iam;
  int ProFlag=FALSE;
  FILE *fp;
  int FpOk=FALSE;
  iam = mynode();
  if(!iam) { if(fp=fopen("out.tex","w")) FpOk=TRUE; }

  if(!iam) 
  { readvalue(&MatrixDimen,&HyperDimen,&px,&py);
    printf("\n MatrixDimen = %d, HyerDimen = %d ",MatrixDimen,HyperDimen);
    printf("\n px = %d, py = %d \n",px,py);
    if(produce_matrix(&A,&B,&C,MatrixDimen)) ProFlag=TRUE;
    sprintf(smsg,"%d",px*py);
    maxnodes=px*py;
    csend(100,smsg,strlen(smsg)+1,-1,0);
    /* send value of ProFlag */
    sprintf(smsg,"%d",ProFlag);
    csend(100,smsg,strlen(smsg)+1,-1,0);
    sprintf(smsg,"%d",MatrixDimen);
    csend(100,smsg,strlen(smsg)+1,-1,0);
    sprintf(smsg,"%d",HyperDimen);
    csend(100,smsg,strlen(smsg)+1,-1,0);
    sprintf(smsg,"%d",px);
    csend(100,smsg,strlen(smsg)+1,-1,0);
    sprintf(smsg,"%d",py);
    csend(100,smsg,strlen(smsg)+1,-1,0);
    if(ProFlag)
    { /*printf("\nMatrix A\n");
      print_matrixS(A,MatrixDimen,MatrixDimen);*/
      /*printf("\nMatrix B\n");
      print_matrixS(B,MatrixDimen,MatrixDimen);*/
      start_time = mclock();
      Matrix_Multiply(A,B,C,MatrixDimen,MatrixDimen,MatrixDimen);
      end_time = mclock();
      /*printf("\n C Matrix \n"); 
      print_matrixS(C,MatrixDimen,MatrixDimen);*/
      printf("\n %d,%d Matrix multiplication, elapsed time : %ld ",MatrixDimen,MatrixDimen,end_time-start_time);
      printf("\n node %d : ProFlag = %s",iam,(ProFlag?"TRUE":"FALSE"));
    }/* end of if */
    if(FpOk) /* If output file is OK! */
    { fprintf(fp,"\n Matrix Size : %d",MatrixDimen);
      fprintf(fp,"\n Hypercube Dimension : %d ",HyperDimen);
      fprintf(fp,"\n Mesh structure is : %d x %d ", py,px);
      fprintf(fp,"\n Total Number of processors : %d ",px*py); 
      /*fprintf(fp,"\n Matrix A \n\n"); 
      Fprint_matrixS(fp,A,MatrixDimen,MatrixDimen);
      fprintf(fp,"\n Matrix B \n\n"); 
      Fprint_matrixS(fp,B,MatrixDimen,MatrixDimen);
      fprintf(fp,"\n Matrix C= A * B \n\n"); 
      Fprint_matrixS(fp,C,MatrixDimen,MatrixDimen);
     */ 
      fprintf(fp,"\n %d,%d Matrix multiplication ",MatrixDimen,MatrixDimen);
      fprintf(fp,", elapsed time : %ld\n ",end_time-start_time);
    }/* end of if */
      printf("\n 2222 node %d : ProFlag = %s",iam,(ProFlag?"TRUE":"FALSE"));
  }
  gsync();
  if(iam)
  { crecv(100,rmsg,sizeof(rmsg));
    maxnodes=atoi(rmsg);
  }

  gsync();
  if((iam)&&(iam<maxnodes))
  { crecv(100,rmsg,sizeof(rmsg));
    ProFlag=atoi(rmsg);
    crecv(100,rmsg,sizeof(rmsg));
    MatrixDimen=atoi(rmsg);
    crecv(100,rmsg,sizeof(rmsg));
    HyperDimen=atoi(rmsg);
    crecv(100,rmsg,sizeof(rmsg));
    px=atoi(rmsg);
    crecv(100,rmsg,sizeof(rmsg));
    py=atoi(rmsg);
  }
  gsync();


  if(ProFlag)
  {  mx=ceil(1.0*MatrixDimen/px);
     my=ceil(1.0*MatrixDimen/py);
     printf("\n mx = %d , my = %d \n",mx,my);
     if(sa = (float *) calloc(mx*my,sizeof(float)))
     { if(sb = (float *) calloc(mx*my,sizeof(float)))
       { if(sc = (float *) calloc(mx*my,sizeof(float)))
         { /* send matrix values to processors */
           start_time = mclock();
           distribute(A,B,MatrixDimen,sa,sb,sc,mx,my,px,py);
           end_time = mclock();
           /*print_matrix(sb,mx,my,px,py);*/
           if(!iam) 
           { printf("\n Distribution takes %ld\n",end_time-start_time);
             if(FpOk)
               fprintf(fp,"\n Distribution takes %ld",end_time-start_time);
           }
            /*print_matrix(sb,mx,my,px,py); */
           
           /*print_matrix(sb,mx,my,px,py);*/ 
           start_time = mclock();
            
           Preconditioning(px,py,MatrixDimen,sa,sb,mx,my,maxnodes);
           end_time = mclock();
           if(!iam) 
           { printf("\n Preconditioning takes %ld\n",end_time-start_time);
             if(FpOk)
                fprintf(fp,"\n Preconditioning takes %ld\n",end_time-start_time);
           }    
           /*print_matrix(sb,mx,my,px,py); */
           start_time = mclock();
           Parallel_Matrix_Multiply(px,py,MatrixDimen,sa,sb,sc,mx,my);
           end_time = mclock();
           if(!iam) 
           { printf("\n Multiplication takes %ld\n",end_time-start_time);
             if(FpOk)
               fprintf(fp,"\n Multiplication takes %ld\n",end_time-start_time);
           }
           /*print_matrix(sc,mx,my,px,py);  Prints Result Matrix */
         }
         else { free(sa);free(sb); }
       }
       else  free(sa);
     }/* end of if */ 
  }
  else
    if(iam<maxnodes)
        printf("\n We do not have enough memory to do this multiplication \n");

  gsync();
  /*iam=mynode();
  for(i=0;i<numnodes();i++)
  { if((iam==i)&&(iam<maxnodes))
           print_matrix(sc,mx,my,px,py);  
    gsync();
  }  
*/
  if(iam<maxnodes)
  {  free(sa);
     free(sb);
     free(sc);
     if(!iam)
     { free(A);
       free(B);
       free(C);
       if(FpOk) fclose(fp);
     }/* end of if */
     if(!iam) printf("\n end of program\n");
  }/* end of if */
  
}/* end of main() */








